/** Open-relay teszter (openrt) by palacsint
*	http://palacsint.hu/
*
*	License: GNU/GPL v2
*	Language: hungarian
*
*	Semmilyen felelosseget nem vallalok a program hasznalata miatt.
*
*
*	2006. aprilis   
*/

#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/types.h>
#include <ctype.h>

#include <regex.h>
#include <locale.h>

#include <errno.h>

/** hlzati hiba */
#define ERROR_NETWORK 3
/** Elrontott paramterezs a felhasznltl */
#define ERROR_PARAM 4
/** programozsi hiba, vagy standard fggvnyek hibi */
#define ERROR_CODING 5

#define RET_OPENRELAY 1
#define RET_NOT_OPEN_RELAY 0

struct param_t {
	int help; // "-h" paramter
	int verbose; // "-v" paramter - verbose uzemmod
	char *host;
	char *from;
	char *to;
	char *comment;
	int timeout;
	int port;
};

int mailaddr_test(char *);

/** Paramterek kinyerse, adatstruktura feltltse
*
*	A program hvsakor megadott paramtereknek megfelel rtkeket lltja be
*	az adatstuktrban, valamint ellenrzi is azokat.
*/
void get_params(int argc, char *argv[], struct param_t *p) {
	int c;
	opterr = 0;

	/** Lehetsges paramtereink valnak... */
	while ((c = getopt (argc, argv, "hvf:u:c:t:")) != -1) {
		switch (c) {
			case 'h':
				p->help = 1;
				break;
			case 'v':
				p->verbose = 1;
				break;
			/** tesztlevl feladja */
			case 'f':
				p->from = optarg;
				if (0 != mailaddr_test(p->from)) {
					fprintf(stderr, "Invalid sender address. (%s)\n", p->from);
					exit(ERROR_PARAM);
				}
				break;
			/** tesztlevl cmzettje */
			case 'u':
				p->to = optarg;
				if (0 != mailaddr_test(p->to)) {
					fprintf(stderr, "Invalid recipient address.\n");
					exit(ERROR_PARAM);
				}
				break;
			/** Tovbbi szveg a tesztlevlbe az alaprtelmezetten kvl */
			case 'c':
				p->comment = optarg;
				break;
			// TODO - ez hasznalva van?
			case 't':
				p->timeout = atoi(optarg);
				if (p->timeout < 0 || p->timeout > 60) {
					fprintf(stderr, "Invalid timeout: %s (interpreted as %d)\n",
						optarg, p->timeout);
					exit(ERROR_PARAM);
				}
				break;
			/** Ismeretlen opci */
			case '?':
				if (isprint (optopt)) {
					fprintf(stderr, "Unknown option `-%c'.\n", optopt);
				} else {
					fprintf(stderr, "Unknown option character `\\x%x'.\n",
						optopt);
				}
				exit(ERROR_PARAM);
				break;
			default:
				fprintf(stderr, "Unknown getopt() action.\n");
				exit(ERROR_CODING);
		}
	}
	
	/** Ha maradt mg feldolgozatlan paramter, akkor az a host lesz -
	*	illetve muszj annak lennie.
	*/
	if (optind < argc) {
		p->host = argv[optind];
	} else {
		fprintf(stderr, "Empty hostname.\n");
		exit(ERROR_PARAM);
	}
	 

	/** Ha mg mindig van paramter, akkor annak a port-nak kell lennie - ez
	*	mr csak opcionlis. Eldntjk, hogy int-e, illetve megfelel-e az 
	*	rtke. Ha nincs port, akkor marad a default (default_params() lltja).
	*/
	optind++;
	if (optind < argc) {
		int tp;
		tp = atoi(argv[optind]);
		if ((tp < 1) || (tp > 65535)) {
			fprintf(stderr, "Invalid port number: %s (interpreted as %d)\n", 
				argv[optind], tp);
			exit(ERROR_PARAM);
		} else {
			p->port = tp;
		}
	}

	/** Ha mg mindig van paramter... az mr neknk tutira nem kell, valamit
	*	elrhatott a felhasznl. 
	*/
	optind++;
	if (optind < argc) {
		fprintf(stderr, "Unknown parameter: %s \n", argv[optind]);
		exit(ERROR_PARAM); 
	}

	return;
}

/** Alapretlmezett adatok bersa az adatstruktrba
*/
void default_params(struct param_t *p) {
	memset(p, 0, sizeof(struct param_t));
	
	p->help = 0;
	p->verbose = 0;
	p->from = p->to = p->comment = NULL; 
	p->timeout = 60;
	p->port = 25; // default SMTP port
	return;
}

/** Verbose zenetek kirsa a kpernyre
*
*	Tobb sorbl ll zenetket is kezel, minden sor elrakja a >>> <<<
*	stringek valamelyikt a from paramter rtktl fggen (C, S)
*	@param	from	char	Kld: 'S', vagy 'C'
*	@param	msg_text	char	Kirand zenet
*/
void print_message(char from, char* msg_text) {
    const char delimiters[] = "\n";
    char *token, *cp;
	int i = 0;
	
    cp = strdup(msg_text); // kell egy msolat az strtok miatt
	if (NULL == cp) {
		fprintf(stderr, "There is not enough memory (strdup).");
		exit(ERROR_CODING);
	}

	/** A \r -ekre nincs szksgnk, csak bekavarn az strtok-ot */	
	while (cp[i] != '\0') {
		if ('\r' == cp[i]) {
			cp[i] = ' ';
		}
		i++;
	}

	/** tokenekre bontjuk, soronkent, majd soronkent kiirjuk */
	token = strtok(cp, delimiters);
	if (NULL == token) {
		fprintf(stderr, "Invalid data? (print_message)\n");
	} else {
		do  {
			if ('S' == from) {
				printf("<<<");
			} else {
				printf(">>>");
			}
			printf(" %s\n", token);
		} while (NULL != (token = strtok(NULL, delimiters)));
	}
	free(cp);
}


/** zenet kldse az SMTP szervernek 
*	@param	sock	int	Kapcsolat-azonost
*	@param	msg_text	char*	Az elkldend uzenet
*	@param	verbose	int	Uzenet kirsa az stdin-re (1: igen, 0: nem)
*	@return	int	Hiba eseten -1, egybknt 0.
*/
int send_message(int sock, char *msg_text, int verbose) {
    
    if (1 == verbose) print_message('C', msg_text);
	
	if (send(sock, msg_text, strlen(msg_text), 0) < 0) {
		perror("send");
		return -1;
    }
    return 0;
}

/** Kilps a programbl s (igny szerint) a kapcsolat lezrsa
*	@param	sock	int	Kapcsolatazonost. Ha nem -1, akkor lezrja close()-al.
*	@param	ret	int	Visszatrsi rtk.
*	@param	verbose	int	Verbose-md (1: igen, ms: nem)
*/
int rt_quit(int sock, int ret, int verbose) {

	if (-1 != sock) {
		send_message(sock, "\r\nQUIT\r\n", verbose);
		close(sock);
	}

	if (1 == verbose) {
		printf("return code: %d\n", ret);
		switch (ret) {
			case ERROR_NETWORK:
				printf("Network error.\n");
				break;
			case ERROR_PARAM:
				printf("Error around the paramter(s).\n");
				break;
			case ERROR_CODING:
				printf("Internal error.\n");
				break;
			case RET_OPENRELAY:
				printf("The host is an open-relay server.\n");
				break;
			case RET_NOT_OPEN_RELAY:
				printf("The host is NOT an open-relay server.\n");
				break;
			default:
				printf("Unknown return value.\n");
				break;
		}
	}
	
	exit(ret);	
}

/** zenet vtele az SMTP szervertl (s ellenrzse)
*	@param sock	int	kapcsolatazonosito
*	@param	answer_ok	int	Ilyen valaszt varunk a szervertol. Ha nem ilyet mond,
*	akkor ERROR_NETWORK koddal kilepunk. Ha 0, akkor figyelmen kivul hagyjuk es
*	a fuggveny visszateresi erteke lesz az SMTP szerver altal kuldott kod.
*	@param	msg_text	char*	Ide kerl a vett vlasz
*	@param	msg_len	int	Ennyi hely van a vlasznak foglalva
*	@param	timeout	int Idtllps (msodperc)
*	@param	verbose int	Verbose-md (1: igen, ms: nem)
*/
int receive_message(int sock, int answer_ok, char *msg_text, int msg_len, int timeout, int verbose) {
	int rcvlen;
	char msg_num[8] = {0}; // ideiglenes hely az uzenet kodjanak
	int msg_code = 0;
	int i;
	
	struct timeval tv; //timeout
	fd_set read_fds; // fajlleiro halmaz
	int fdmax;
	int retval;

	FD_ZERO(&read_fds); 
	FD_SET(sock, &read_fds); 
	fdmax = sock;

	tv.tv_sec = timeout; 
	tv.tv_usec = 0;

	/** Csak az idtllps miatt van select() */
	retval = select(fdmax + 1, &read_fds, NULL, NULL, &tv);
	if (-1 == retval) {
		perror("select");
		rt_quit(sock, ERROR_NETWORK, verbose);
	}

	if (FD_ISSET(sock, &read_fds)) {
		rcvlen = recv(sock, msg_text, msg_len, 0); // uzenet vetele		
	} else {
		rt_quit(sock, ERROR_NETWORK, verbose);
	}
	
	if (rcvlen < 0) {
		perror("recv");
		rt_quit(sock, ERROR_NETWORK, verbose);
	} else if (0 == rcvlen) {
		fprintf(stderr, "Connection reset by peer.\n");
		rt_quit(-1, ERROR_NETWORK, verbose);
	} else {
		msg_text[rcvlen] = '\0';
		if (1 == verbose) print_message('S', msg_text);
		
		if (NULL != (strchr(msg_text, ' '))) {
			/** Megkeresi az elso szokozt az uzenetben. Elotte lesz a kod,
			ami nekunk kell */
		    i = 0;
		    while ((msg_text[i] != ' ') && (msg_text[i] != '\0')) {
				i++;
				// ettol hosszab hibakod nem nagyon van
				if (i > 5) {
					fprintf(stderr, "Invalid answer from the SMTP server.\n");
					rt_quit(sock, ERROR_NETWORK, verbose);
				}
		    }

			i = (i > sizeof(msg_num) - 1) ? sizeof(msg_num) - 1 : i;
		    strncpy(msg_num, msg_text, i);
		    //printf("num: %s\n", msg_num);
		    msg_code = atoi(msg_num);
			
			/** Ervenyes-e az uzenetkod? */
			if (msg_code > 999 || msg_code < 100) {
				fprintf(stderr, "Invalid answer from the SMTP server.\n");
				rt_quit(sock, ERROR_NETWORK, verbose);
			}

			/** Ha ilyen kapunk, akkor tuti nem open-relay, ki is lephetunk */
			/** Talan nem a legszebb, de hatekony */
			if (550 == msg_code || 554 == msg_code) {
				rt_quit(sock, RET_NOT_OPEN_RELAY, verbose);
			}

			if ((msg_code != answer_ok) && (0 != answer_ok)) {
				fprintf(stderr, "Invalid answer from the SMTP server.\n");
				if (1 != verbose) {
					print_message('S', msg_text);
				}
				rt_quit(sock, ERROR_NETWORK, verbose);
			}
			
		} else {
		    fprintf(stderr, "Invalid answer from the SMTP server.\n");
		    rt_quit(sock, ERROR_NETWORK, verbose);
		}
	}
	
	return msg_code;
}

/** Hasznalati utasitas megjelenitese
*/
void usage(void) {
	printf("Usage:\topenrt [-v] [-t timeout] -f sender_addr "); 
	printf("-u recipient_addr [-c comment] <hostname> <port>\n");
	printf("\topenrt -h\n");
	printf("Options:\n \
	-h\t\t\tthis help\n  \
	-f sender_addr\t\tsender-addr\n  \
	-u recipient_addr\tto-addr\n  \
	-v\t\t\tverbose mode\n  \
	-c comment\t\textra comments to the mail\n  \
	-t timeout\t\ttimeout in second(s)\n");
}

/** SMTP szerver tesztelese
*	@param	sock	int	Mar felepitett kapcsolat azonositoja
*	@param	params	struct param_t* A parancssorban megadott paramtereket
*	feldolgozva tartalamz tmb
*	@return	int
*/
int test_smtp(int sock, struct param_t *params) {
	int ret = 0;
	char msg_text[512];
	char tmp_text[2048];
	char host[256];
	
	/** Udvozlobanner fogadasa */
	ret = receive_message(sock, 220, msg_text, sizeof(msg_text), params->timeout, params->verbose);

	/** Helyi gep nevenek meghatarozasa a koszoneshez  */
	if (-1 == gethostname(host, sizeof(host))) {
		fprintf(stderr, "Couldn't get hostname (gethostname()).\n");
		exit(ERROR_CODING);
	}
	
	/** Koszonunk, HELO kuldese
	Az EHLO mr advanced, s arra tbb sor is jhet vlaszknt (a HELO-ra nem).
	Az RFC szerint minden SMTP-nek tudnia kell a HELO-t is (s tudjk is)
	*/
	ret = snprintf(tmp_text, sizeof(tmp_text), "HELO %s\r\n", host);
	if (ret < 0) {
		fprintf(stderr, "Internal error (snprintf).");
		exit(ERROR_CODING);
	}
	send_message(sock, tmp_text, params->verbose);
	
	/** 250 OK-ot varunk */
	receive_message(sock, 250, msg_text, sizeof(msg_text), params->timeout, params->verbose);

	/** Fejlec - felado */
	ret = snprintf(tmp_text, sizeof(tmp_text), "MAIL FROM: %s\r\n", params->from);
	if (ret < 0) {
		fprintf(stderr, "Internal error (snprintf).");
		return ERROR_CODING;
	}

	send_message(sock, tmp_text, params->verbose);
	receive_message(sock, 250, msg_text, sizeof(msg_text), params->timeout, params->verbose);

	/** Itt szokott 550 lenni */
	ret = snprintf(tmp_text, sizeof(tmp_text), "RCPT TO: %s\r\n", params->to);
	if (ret < 0) {
		fprintf(stderr, "Internal error (snprintf).");
		return ERROR_CODING;
	}
	
	send_message(sock, tmp_text, params->verbose);
	ret = receive_message(sock, 0, msg_text, sizeof(msg_text), params->timeout, params->verbose);

	
	send_message(sock, "DATA\r\n", params->verbose);
	receive_message(sock, 354, msg_text, sizeof(msg_text), params->timeout, params->verbose);

	/** Ez lesz az uzenet */
	ret = snprintf(tmp_text, sizeof(tmp_text), "To: %s\r\n" \
			"From: %s\r\n" \
			"Subject: open-relay test message\r\n" \
			"\r\n" \
			"This is a test message which checks your SMTP server as an\r\n" \
			"open-relay server and tries to send a message with its.\r\n" \
			"\r\n" \
			"Please ignore this e-mail.\r\n" \
			"\r\n" \
			"%s\r\n" \
			"\r\n" \
			"EOF\r\n" \
			".\r\n", params->to, params->from, params->comment);
	if (ret < 0) {
		fprintf(stderr, "Internal error (snprintf).");
		return ERROR_CODING;
	}
			
	send_message(sock, tmp_text, params->verbose);
	ret = receive_message(sock, 0, msg_text, sizeof(msg_text), params->timeout, params->verbose);
	
	if (250 == ret) {
		ret = RET_OPENRELAY;
	} else {
		/** Valami oknal fogva megsem sikerult elkuldeni a levelet */
		ret = RET_NOT_OPEN_RELAY;
	}

	return ret;
}

/** Mailcim ervenyessegenek tesztelese (regexp alapjan)
*
*	@param mailaddr char*	Tesztelendo e-mail cim
*	@return	int	0: OK, egyeb: hiba
*/
int mailaddr_test(char *mailaddr) {
	int ret;
	regex_t  ford;
	
	ret = 0; // visszateresi ertek

	/** Ez biztos nem jo e-mail cim */
	if (NULL == mailaddr) {
		return 1;
	}

	/** Egyszeru regexp, a hibak nagy reszet kiszuri */
    if (regcomp(&ford, 
		"^[_A-Za-z0-9-]+(\\.[_A-Za-z0-9-]+)*@[A-Za-z0-9-]+(\\.[A-Za-z0-9-]+)*$", 
		REG_EXTENDED) != 0) {
			/** Itt mar komoly gondok vannak... */
			fprintf(stderr, "Internal error in the regular expression.\n");
			exit(ERROR_CODING);
	}
	
	if ((ret = regexec(&ford, mailaddr, 0, NULL, 0)) == 0) {
		ret = 0; // OK az e-mail cim
	} else {
		ret = 1; // nem illeszkedik, hibas
	}

	regfree(&ford);
	return ret;
}

/** main() - szokasos argumentumok...
*/
int main(int argc, char *argv[]) {
	struct hostent *host;
	struct sockaddr_in addr;
	
	int ret; // visszateresi ertek
	int sock; // socket azonosito

	struct param_t params;
	
	/** Rootkent inkabb ne hasznaljuk... */
	if ((0 == geteuid()) || (0 == getegid())) {
		fprintf(stderr, "Don't use this program as root.\n");
		exit(ERROR_PARAM);
	}

	/** Ha nem adott meg parametert, akkor sugo */	
	if (2 > argc) {
		usage();
		exit(0);
	}
	
	/** megadott parameterek kezelese */
	default_params(&params);
	get_params(argc, argv, &params);
	
	if (1 == params.help) {
		usage();
		exit(0);
	}
	
	if (NULL == params.host || NULL == (host = gethostbyname(params.host))) {
		fprintf(stderr, "Invalid hostname.\n");
		exit(ERROR_PARAM);
	}
	
	if (0 != params.verbose) {
		printf("Connection to %s...\n", params.host);
	}
	
	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		exit(ERROR_NETWORK);
	}
	
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(params.port);
	addr.sin_addr = *((struct in_addr *) (host->h_addr));
	
	if (connect(sock, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		perror("connect");
		exit(ERROR_NETWORK);
	}

	if (0 != params.verbose) {
		printf("Connection established.\n");
	}

	/** Teszt */
	ret = test_smtp(sock, &params);

	/** Kapcsolat lezeras es kilepes */
	rt_quit(sock, ret, params.verbose);

	return 0; // ez itt mar nem is nagyon kellene
}
